# -*- coding: utf-8 -*-
""" `lunatic.pool` module.

Provides a high level connection pooling for Postgresql servers.
"""


__all__ = ('DBEngine', 'DBRouter', 'DBEngineError', 'DBRouterError')


import logging
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool, QueuePool
from itertools import cycle
from sqlalchemy.exc import ProgrammingError, OperationalError, DatabaseError, DataError
from lunatic.errors import DBEngineError, DBRouterError


class DBEngine(object):
    """Postgresql database connection engine class.

    Provides a wrapper on sqlalchemy DBAPI-2 implementation implementing
    a concrete and simple API.

    Attributes:
        conn_uri (str): PostgreSQL connection info.
        pool_type (object): SQLAlchemy pool factory class.
        engine (object): SQLAlchemy Engine instance.
        debug (bool): Set DEBUG mode on / off.


    Raises:
        DBEngineError, if any exception occurs.
    """

    __slots__ = ('conn_uri', 'pool_type', '_engine', 'debug')

    logger = logging.getLogger(__name__)

    pool_factory = (
        ('null', NullPool),
        ('queue', QueuePool)
    )

    def __init__(self, conn_uri='', pool_factory='null', debug=True, **conn_data):
        self.conn_uri = conn_uri if conn_uri else self.pg_uri(**conn_data)
        self.pool_type = dict(self.pool_factory).get(pool_factory, NullPool)
        self.debug = debug
        self._engine = None

    @property
    def engine(self):
        """Lazy initialize DBAPI-2 connection protocol.
        """
        if not self._engine:
            self._engine = create_engine(
                self.conn_uri,
                poolclass=self.pool_type,
                isolation_level='AUTOCOMMIT'
            )

            self._engine._use_thread_local = True

            if self.debug:  # pragma: no cover
                self.logger.info('Initialized pool: {}'.format(self.conn_uri))

        return self._engine

    def query(self, qs, fetch_many=True):
        """Execute query within engine.

        Args:
            qs (str): Plain text query.
            fetch_many (boolean)
        Returns:
            Query Records or None.

        Raises:
            DBEngineError, if a connection / data Error occurs.
        """
        try:
            record_proxy = self.engine.execute(qs)

        except (ProgrammingError, OperationalError, DatabaseError, DataError) as error:
            self.logger.debug('Error Encountered: {}'.format(error.args[0]))
            raise DBEngineError(error.args[0])

        records = [{key: row[key] for key in record_proxy.keys()} for row in record_proxy]

        if self.debug:  # pragma: no cover
            self.logger.info('SQL: {}'.format(qs))

        if fetch_many:
            return records or []

        return records[0] if records else None

    @staticmethod
    def pg_uri(host='localhost', port=5432, database='postgres', user='postgres', password=None):
        """Return a PostgreSQL connection URI for the specified values.

        Args:
            host (str): Host to connect to
            port (int): Port to connect on
            database (str): The database name
            user (str): User to connect as
            password (str): The password to use, None for no password

        Returns:
            A valid PostgreSQL connection URI.
        """
        if port:
            host = '{}:{}'.format(host, port)

        if password:
            return 'postgresql://{}:{}@{}/{}'.format(user, password, host, database)

        return 'postgresql://{}@{}/{}'.format(user, host, database)

    def __repr__(self):  # pragma: no cover
        return '<DBEngine instance at: 0x{:x}>'.format(id(self))

    def __str__(self):  # pragma: no cover
        return '<DBEngine({})>'.format(self.conn_uri)

    def __call__(self, qs, fetch_many=True):  # pragma: no cover
        return self.query(qs, fetch_many)


class DBRouter(object):
    """DBEngine Router class.

    Manages multiple DBEngine instances. `DBRouter` instances
    implements round robin load balance protocol.

    Attributes:
        _engines (dict): a mapping of aliases / DBEngine instances for management.
    """

    logger = logging.getLogger(__name__)

    __slots__ = ('_engines', 'router', 'debug')

    def __init__(self, debug=True, **engines):
        self._engines = engines or {}
        self.debug = debug
        self.router = cycle(engines.keys())

    @classmethod
    def dict_config(cls, **config):
        """Initialize DBRouter from a dictionary/

        Args:
            config (): A mapping of aliases / DBEngine initialization parameters.

        Returns:
            DBRouter instance.
        """
        return cls(**{key: DBEngine(**config[key]) for key in config})

    @property
    def engines(self):
        return tuple(self._engines.keys())

    def proxy_query(self, qs, fetch_many=True):
        """Execute query within an managed engine.

        Args:
            qs (str): Plain text query.
            fetch_many (boolean)
        Returns:
            Query Records or None.

        Raises:
            DBEngineError, if a connection / data Error occurs.
        """

        db_route = next(self.router)

        if self.debug:  # pragma: no cover
            self.logger.info('Routing query to: {}'.format(db_route))
            return self._engines.get(db_route).query(qs, fetch_many)

    def __getitem__(self, item):
        return self._engines[item]

    def __contains__(self, item):
        return item in self.engines

    def __setitem__(self, key, value):
        if not isinstance(value, DBEngine):
            raise DBRouterError('DBRouter accepts only `lunatic.DBEngine` instances.')
        self._engines[key] = value
        self.router = cycle(self.engines)

    def __delitem__(self, key):
        del self._engines[key]
        self.router = cycle(self.engines)



# -*- coding: utf-8 -*-
"""`lunatic.query` module.

Provides a query interface for interacting with `lunatic.engine.DBEngine`
objects.
"""
import logging
from lunatic.errors import DBEngineError, DBRouterError, QueryManagerError
from snaql.factory import Snaql, SnaqlException
from functools import wraps


class QueryManager(object):
    """Base query manager class.
    """

    debug = True
    logger = logging.getLogger(__name__)

    def __init__(self, engine=None):
        self.engine = engine
        self.builder = None

    def load(self, query_parent, query_dir, query_file):
        """Load queries from template files to instance attributes.

        Args:
            query_parent (str): Query folder parent directory
            query_dir (str): Query folder name
            query_file (str): Query file name.

        Returns:
            The QueryManager instance.

        Raises:
            QueryManagerError, if any snaql Exception occurs.
        """
        try:
            self.builder = Snaql(query_parent, query_dir).load_queries(query_file)
        except SnaqlException as error:
            raise QueryManagerError(error.args[0])

        self.mount()

        return self

    @property
    def proxy_rule(self):
        return [method for method in dir(self.builder)
                if not method.startswith('_')
                and hasattr(getattr(self.builder, method), 'func_name')]

    def mount(self):
        """Mount snaql queries instance.
        """

        def proxy_wrapper(func):
            @wraps(func)
            def _wrapper(*args, **kwargs):
                """Inner wrapper for proxy methods.
                """
                fetch_many = kwargs.pop('fetch_many', True)

                try:
                    queryset = func(*args, **kwargs)
                    if self.debug:
                        self.logger.info('Execute: {}'.format(queryset))
                    records = self.engine.query(queryset, fetch_many)

                except (SnaqlException, DBRouterError, DBEngineError) as error:
                    self.logger.error(error.args[0])
                    raise QueryManagerError(error.args[0])

                return records

            return _wrapper

        for proxy in self.proxy_rule:
            setattr(self, proxy, proxy_wrapper(getattr(self.builder, proxy)))

        return self




# -*- coding: utf-8 -*-
#
#    Copyright (C) 2015  Papavassiliou Vassilis
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""`lunatic` package.

Postgresql python booster package!
"""

__all__ = ('DBEngine', 'DBEngineError', 'DBRouter', 'DBRouterError', 'QueryManager', 'QueryManagerError',
           'pubsub', 'patch')

__version__ = '1.1.4'

from lunatic.engine import DBEngine, DBRouter
from lunatic.query import QueryManager
import lunatic.pubsub
import lunatic.patch
from lunatic.errors import (DBRouterError, DBEngineError, LunaticError, QueryManagerError)
import logging


logging.getLogger('lunatic').addHandler(logging.NullHandler)



# -*- coding: utf-8 -*-
"""`lunatic.errors` module.

Provides package base error classes.
"""


class LunaticError(Exception):
    """Base package Error class.
    """
    pass


class QueryManagerError(LunaticError):
    """QueryManager Error class.
    """
    pass


class DBEngineError(LunaticError):
    """DBEngine base Error.
    """
    pass


class DBRouterError(DBEngineError):
    """DBRouter Error.
    """
    pass



# -*- coding: utf-8 -*-
"""`pgtools.pubsub` module.

Provides a nice API over Postgresql LISTEN / NOTIFY.
"""

__all__ = ('connect', )

import select
import logging

import psycopg2

logger = logging.getLogger(__name__)

NOT_READY = ([], [], [])


class PubSub(object):
    """Basic pub-sub handler class.
    """
    def __init__(self, conn):
        assert conn.autocommit, "Connection must be in autocommit mode."
        self.conn = conn

    def listen(self, channel):
        with self.conn.cursor() as cur:
            cur.execute('LISTEN %s;' % channel)

    def unlisten(self, channel):
        with self.conn.cursor() as cur:
            cur.execute('UNLISTEN %s;' % channel)

    def notify(self, channel, payload):
        with self.conn.cursor() as cur:
            cur.execute('SELECT pg_notify(%s, %s);', (channel, payload))

    def get_event(self, select_timeout=0):
        # poll the connection, then return one event, if we have one.  Else
        # return None.
        select.select([self.conn], [], [], select_timeout)
        self.conn.poll()
        if self.conn.notifies:
            return self.conn.notifies.pop(0)

    def get_events(self, select_timeout=0):
        # Poll the connection and return all events, if there are any.  Else
        # return None.
        select.select([self.conn], [], [], select_timeout)  # redundant?
        self.conn.poll()
        events = []
        while self.conn.notifies:
            events.append(self.conn.notifies.pop(0))
        if events:
            return events

    def events(self, select_timeout=5, yield_timeouts=False):
        while True:
            if select.select([self.conn], [], [], select_timeout) == NOT_READY:
                if yield_timeouts:
                    yield None
            else:
                self.conn.poll()
                while self.conn.notifies:
                    yield self.conn.notifies.pop(0)

    def close(self):
        self.conn.close()


def connect(*args, **kwargs):
    conn = psycopg2.connect(*args, **kwargs)
    conn.autocommit = True
    return PubSub(conn)



# -*- coding: utf-8 -*-
"""`pgtools.dbapi` module.

Provides models and utilities for using Postgresql Database API's
(UDF/Views) with parameter validation.
"""

__license__ = 'GLPv3'
__date__ = '1-9-015'
__author__ = 'pav'

__all__ = ['DBAPIBackend', 'FunctionField', 'ViewField', 'DBAPIError', 'UnknownParamError',
           'InvalidFunctionParamError']

# Try to find the best candidate for JSON serialization.
# Import order indicates serialization efficiency.


try:
    import ujson as json
except ImportError:
    try:
        import simplejson as json
    except ImportError:
        import json


class DBAPIError(Exception):
    """Base module exception
    """
    pass


class InvalidFunctionParamError(DBAPIError):
    """Raises when invalid type parameter is passed to Postgresql Function.
    """
    pass


class BaseFuncArg(object):
    """**Base Postgresql Function parameter validator**

    Implements a base validator template class for function parameters.
    Subclass it and define the attributes.


    Attributes:
        - formatter (str): The string template corresponding to DB native form.
        - param_type (tuple): The parameter native classes allowed.
    """

    formatter = ''
    param_type = None

    def __init__(self, param):
        self.param = param

    def validate(self):
        if not isinstance(self.param, self.param_type):
            raise InvalidFunctionParamError(
                "Invalid type for {} param".format(self.param_type)
            )
        return self.formatter.format(
            self.format()
        )

    def format(self):
        return self.param


class BaseAPIField(object):
    """**Base ModelField Descriptor class**

    Model Field Descriptor Abstract class. It provides the minimum interface
    for building Model class attributes that map to Database object. It is not
    indented to be initialized, although won't raise a ``TypeError``.

    You need to subclass it and override at least ``__get__`` special method
    or add any method you like. BaseField encapsulates an inner functionality
    that subclasses need. For now its only ``auto-tagging``.

    Attributes:
        - field (str): The descriptor name for instance owner class.

    .. note::
        BaseField will also implement ``lazyload`` functionality. So if you
        really need to override ``__init__`` special method don't forget to
        call ``super`` (as shown in example below).

    Subclass example::

        >>> class MyCustomField(BaseAPIField):
        ...     def __init__(self, my_value, **kwargs):
        ...         self.custom_attr = my_value
        ...         super(MyCustomField, self).__init__(**kwargs)
        ...
        ...     def __get__(self, instance, owner):
        ...         return instance.some_attr is self.custom_attr
        ...

    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.field = None

    def __get__(self, instance, owner):
        """Implementing descriptor '__get__' method.
        """
        raise NotImplementedError


class NumArg(BaseFuncArg):
    """Numeric Parameters formatter/validator class.
    """
    formatter = '{}'
    param_type = (int, float)


class TextArg(BaseFuncArg):
    """Textual Parameters formatter/validator class.
    """
    formatter = "'{}'"
    param_type = (str, )


class BooleanArg(BaseFuncArg):
    """Boolean Parameters formatter/validator class.
    """

    formatter = '{}'
    param_type = (bool, )

    def format(self):
        return str(self.param)


class ArrayArg(BaseFuncArg):
    """Array Parameters formatter/validator class.
    """
    formatter = "'{}'"
    param_type = (list, tuple)

    def format(self):
        array_data = str(self.param)[1:-1].replace("u'", "").replace("'", '')
        return '{%s}' % array_data


class JSONArg(BaseFuncArg):
    """JSON Parameters formatter/validator class.
    """
    formatter = "'{}'"
    param_type = (dict, )

    def format(self):
        return json.dumps(self.param)


FUNC_TYPES = {
    int: NumArg,
    float: NumArg,
    str: TextArg,
    dict: JSONArg,
    bool: BooleanArg,
    list: ArrayArg,
    tuple: ArrayArg
}


class UnknownParamError(DBAPIError):
    """Raises when passing unknown parameter in Postgresql function.
    """
    pass


def param_validator(param_specs, params, order=None):
    """Parameter validator function.

    >>> param_specs = {"email": str, "tags": list}
    >>> params = {"email": "pav@gmail.com", "tags": [1, 2, 3, 4, 5]}
    >>> print(', '.join(sorted(param_validator(param_specs, params))))
    'pav@gmail.com', '{1, 2, 3, 4, 5}'
    """
    if not set(param_specs).issuperset(set(params)) or not params:
        raise UnknownParamError("Invalid function arguments: {} - {}".format(
            set(param_specs), set(params)
        ))

    if not order:
        return (FUNC_TYPES[param_specs[key]](params[key]).validate()
                for key in params)

    return (FUNC_TYPES[param_specs[key]](params[key]).validate()
            for key in order)


class ViewField(BaseAPIField):
    """**ViewField Descriptor class**

    ViewField is non-data Descriptor that subclasses
    :class:`pgtools.dbapi.BaseAPIField` and wraps a Database schema view object
    .It should be used with no initial arguments at the meantime (as show in
    example below).

    .. note::
        ViewField class can be initialized with arguments, but so far they will
        have no effect. The only reason that it is validated is that soon you
        could pass parameters that control ``cache_property`` behavior and
        result pagination. Classes with ``ViewField`` attributes *must* have
        a `persistence` attribute that implements
        `pg_utils.bases.BaseConnectionPool`interface.

    Example usage::

        >>> class MyClass(object):
        ...     test_function = ViewField()
        ...
    """

    def __get__(self, instance, owner):
        """Implementing descriptor '__get__' method.
        """
        if not instance and owner:
            return self

        def _callback():
            return 'SELECT * FROM {};'.format(self.field)

        return _callback


class FunctionField(BaseAPIField):
    """**FunctionField Descriptor class**

    FunctionField is also a non-data descriptor class that subclasses
    :class:`pgtools.dbapi.BaseAPIField` for wrapping database schema functions.

    You must initialize it with python native types that
    correspond to Database function parameter types (as you example below).
    The ``__get__`` method returns a callable object, that acts as the actual
    database function with parameter validation.

    .. note::
        FunctionField can also be initialized with extra parameters with no
        effect for the meantime. When all functionality is enabled you should
        explicitly declare kwargs, for easier debugging. The class that has
        ``FunctionField`` attributes *must* have a `persistence` attribute
        that implements `pg_utils.bases.BaseConnectionPool`interface.


    Example Usage (assuming there is a db function with that name and param)::

        >>> class MyModel(object):
        ...     get_model_by_pk = FunctionField(pk=int)
        ...
    """

    def __init__(self, order=None, **func_params):
        self.order = order
        self.func_specs = func_params
        super(FunctionField, self).__init__(**func_params)

    def __get__(self, instance, owner):
        """Implementing descriptor '__get__' method.
        """
        if not instance and owner:
            return self

        def func_callable(**args):

            function_query = 'SELECT * FROM {}({});'.format(
                self.field,
                ', '.join(param_validator(
                    self.func_specs, args, self.order
                ))
            )

            return function_query

        return func_callable


class DBAPIMeta(type):
    """Base metaclass for DBAPIBackend classes.

    ModelMeta ensures two things.

    1] Initialized Classes implement `Singleton` pattern.
    2] ViewFields, and FunctionFields auto tagging from class
       labeling.
    """
    _instances = {}

    def __new__(mcs, name, bases, attrs):

        cls_meta = attrs.get('Meta')
        cls_schema = getattr(cls_meta, 'schema', None)
        for cls_tag, attr in attrs.items():
            try:
                if BaseAPIField in attr.__class__.__mro__:
                    attr.field = "{}.{}".format(
                        cls_schema or 'public',
                        cls_tag
                    )
                    attr.tag = cls_tag
            except AttributeError:
                pass
        return super(DBAPIMeta, mcs).__new__(mcs, name, bases, attrs)

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                DBAPIMeta,
                cls
            ).__call__(*args)
        return cls._instances[cls]


class DBAPIBackend(object, metaclass=DBAPIMeta):
    """**Abstract Base DBAPIBackend class**

    This is the base model class for implementing data models that correspond
    to Database namespace objects. Derivatives from BaseModel in order to use

    Subclasses implement by default the *Singleton* pattern (that means that
    only an instance is created) and also enables BaseField ``auto-tagging``
    functionality.

    Example Usage::

        >>> class ProductModel(DBAPIBackend):
        ...     # declare `function` API
        ...     get_product = FunctionField(pk=int)
        ...     create_product = FunctionField(product_name=str)
        ...
        ...     # declare `view` API
        ...     latest_products = ViewField()
        ...     count_products = ViewField()
        ...
        ...     class Meta:
        ...         schema = 'product'
        ...
        >>> product = ProductModel()
        >>> product.count_products()
        'SELECT * FROM product.count_products;'
        >>> product.get_product(pk=6526352)
        'SELECT * FROM product.get_product(6526352);'


    """


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""
"""

__all__ = ('psycopg_to_gevent', 'json_to_ordered_dict')

from collections import OrderedDict
from psycogreen.gevent import patch_psycopg
import json
import psycopg2.extras


def psycopg_to_gevent():
    """`psycogreen.gevent` shortcut.
    """
    patch_psycopg()


def json_to_ordered_dict():
    """Return PostgreSQL JSON/JSONB types to python OrderedDict.
    """
    def jsonb_loads(stream):
        return json.loads(stream, object_pairs_hook=OrderedDict)

    psycopg2.extras.register_default_json(globally=True, loads=jsonb_loads)



from distutils.core import setup
from lunatic import __version__

version = __version__

setup(
    name='lunatic',
    version=version,
    packages=['lunatic'],
    url='https://github.com/agile4you/lunatic',
    requires=[
        'psycopg2',
        'gevent',
        'psycogreen',
        'psycogreen',
        'ujson',
        'sqlalchemy',
        'snaql'
    ],
    license='GLPv3',
    author='Papavassiliou Vassilis',
    author_email='vpapavasil@gmail.com',
    description='PostgreSQL Python booster utilities'
)
